# github_directory: zino-hofmann/graphql-flutter, stars: 2978, last_update: 2022-07-10
from graphdna.detectors.checkers import in_response_text
from graphdna.entities.interfaces.heuristics import IGQLQuery


class Flutter(IGQLQuery):

    score_factor = 0.55
    genetics = {
        'query { __typename @deprecated }': in_response_text('Directive \\"deprecated\\" may not be used on FIELD.'),
    }



"""Handle DNA flow for a given GraphQL endpoint."""

import asyncio
import logging
from typing import Dict, Optional

from graphdna.entities import GraphQLEngine
from graphdna.entities.interfaces import IGraphDNA
from graphdna.heuristics import HeuristicsManager
from graphdna.http import HTTPBucket
from graphdna.logger import setup_logger


def detect_engine(
    url: str,
    headers: Optional[Dict[str, str]] = None,
    logger: Optional[logging.Logger] = None,
) -> Optional[GraphQLEngine]:
    """Manage the engine detection flow."""

    dna = GraphDNA(
        url,
        headers=headers,
        logger=logger,
    )
    return asyncio.run(dna.run())


async def detect_engine_async(
    url: str,
    headers: Optional[Dict[str, str]] = None,
    logger: Optional[logging.Logger] = None,
) -> Optional[GraphQLEngine]:
    """Manage the engine detection flow asyncronously."""

    dna = GraphDNA(
        url,
        headers=headers,
        logger=logger,
    )
    return await dna.run()


class GraphDNA(IGraphDNA):

    """Manage the DNA of the GraphQL endpoint."""

    def __init__(
        self,
        url: str,
        headers: Optional[Dict[str, str]] = None,
        logger: Optional[logging.Logger] = None,
    ) -> None:
        """Init class."""

        self._url = url
        self._logger = logger or setup_logger()
        self._http_bucket = HTTPBucket(
            self._logger,
            self._url,
            headers,
        )
        self._logger.info(f'Initializing GraphDNA for {url}.')

    async def run(self) -> Optional[GraphQLEngine]:
        """Manage DNA test flow."""

        heuristics = HeuristicsManager(self._logger)
        heuristics.load()

        await heuristics.enqueue_requests(self._http_bucket)
        await self._http_bucket.consume_bucket()
        await heuristics.parse_requests(self._http_bucket)
        heuristics.display_results()

        await self._http_bucket.close_session()

        return heuristics.best_candidate



import asyncio
import logging
from typing import AsyncGenerator, Dict, List, Tuple

import aiohttp

from graphdna.entities.engines import GraphQLEngine
from graphdna.entities.interfaces.dna import IHTTPBucket, IRequest
from graphdna.entities.interfaces.heuristics import EvalMethods, IGQLQueriesManager, IGQLQuery
from graphdna.heuristics.utils import find_engine, import_heuristics


class GQLQueriesManager(IGQLQueriesManager):

    def __init__(
        self,
        logger: logging.Logger,
    ) -> None:
        self._heuristics = []
        self._logger = logger

    def load(self) -> None:
        raw_heuritics = import_heuristics(
            __file__,
            __name__,
        )

        heuristics: List[IGQLQuery] = []
        for raw in raw_heuritics:
            engine = raw.__name__.split('.')[-1]
            engine = find_engine(engine, dir(raw))

            cls = eval(f'raw.{engine}')  # pylint: disable=eval-used
            cls.__engine__ = GraphQLEngine(engine)
            heuristics.append(cls)

        self._heuristics = heuristics

    async def enqueue_requests(
        self,
        bucket: IHTTPBucket,
    ) -> None:
        # Refactor enqueue with a generator
        for heuristic in self._heuristics:
            new_genetics: Dict[str, EvalMethods] = {}

            for query, _evals in heuristic.genetics.items():
                req = IRequest('%%url%%', 'POST', {
                    'json': {
                        'query': query,
                    },
                    'headers': {
                        'Content-Type': 'application/json',
                    },
                })
                req_hash = await bucket.put(req)
                new_genetics[req_hash] = _evals

            heuristic.genetics = new_genetics

    async def parse_requests(  # type: ignore[override]
        self,
        bucket: IHTTPBucket,
    ) -> AsyncGenerator[Tuple[IGQLQuery, GraphQLEngine], None]:
        for heuristic in self._heuristics:
            for req_hash, _evals in heuristic.genetics.items():
                client_response = bucket.get(req_hash)
                if not client_response:
                    continue

                if not isinstance(_evals, List):
                    _evals = [_evals]

                for _eval in _evals:
                    try:
                        if not await _eval(client_response):
                            continue
                    except (aiohttp.client_exceptions.ContentTypeError, asyncio.TimeoutError):
                        self._logger.error('Response content unpacking failed.')
                        continue

                    yield heuristic, heuristic.__engine__



from graphdna.detectors.checkers import in_response_text
from graphdna.entities.interfaces.heuristics import IGQLQuery


class Stepzen(IGQLQuery):

    score_factor = 1
    genetics = {
        '': in_response_text('Must provide an operation.'),
    }



"""Exports all heuristics managers that are applied to a graphql endpoint."""

from graphdna.heuristics.gql_queries import GQLQueriesManager  # noqa
from graphdna.heuristics.manager import HeuristicsManager  # noqa
from graphdna.heuristics.web_properties import WebPropertiesManager  # noqa



"""Manage the heuristics interfaces."""

import functools
import logging
from abc import ABC, abstractmethod
from typing import Any, AsyncGenerator, Dict, List, Optional, Tuple, Union

from graphdna.entities.engines import GraphQLEngine
from graphdna.entities.interfaces import IHTTPBucket
from graphdna.entities.interfaces.dna import IRequest

EvalMethods = Union[functools.partial, List[functools.partial]]


class IHeuristic(ABC):

    score: int = 100
    score_factor: float = 1.0

    def verify(self) -> bool:
        raise NotImplementedError


class IGQLQuery(IHeuristic):

    genetics: Dict[str, EvalMethods]


class IWebProperty(IHeuristic):

    requests: List[Tuple[IRequest, EvalMethods]]


class IHeuristicManager(ABC):

    _logger: logging.Logger

    @abstractmethod
    async def enqueue_requests(
        self,
        bucket: IHTTPBucket,
    ) -> None:
        ...

    @abstractmethod
    def load(self) -> None:
        ...

    @abstractmethod
    async def parse_requests(
        self,
        bucket: IHTTPBucket,
    ) -> AsyncGenerator[Tuple[Any, GraphQLEngine], None]:
        ...


class IGQLQueriesManager(IHeuristicManager):

    _heuristics: List[IGQLQuery]


class IWebPropertiesManager(IHeuristicManager):

    _heuristics: List[IWebProperty]


class IHeuristicsManager(ABC):

    _logger: logging.Logger
    _candidates: Dict[GraphQLEngine, int]

    _gql_queries_manager: IHeuristicManager

    @abstractmethod
    def load(self) -> None:
        ...

    @abstractmethod
    async def enqueue_requests(
        self,
        bucket: IHTTPBucket,
    ) -> None:
        ...

    @abstractmethod
    async def parse_requests(
        self,
        bucket: IHTTPBucket,
    ) -> None:
        ...

    @abstractmethod
    def add_score(
        self,
        cls: object,
        engine: GraphQLEngine,
    ) -> None:
        ...

    @abstractmethod
    def display_results(self) -> None:
        ...

    @property
    @abstractmethod
    def best_candidate(self) -> Optional[GraphQLEngine]:
        ...



# github_directory: dgraph-io/dgraph, stars: 18204, last_update: 2022-07-10
from graphdna.detectors.checkers import in_response_text
from graphdna.entities.interfaces.heuristics import IGQLQuery


class DGraph(IGQLQuery):

    score_factor = 0.8
    genetics = {
        'query { __typename }':
            in_response_text([
                'Not resolving __typename. There\'s no GraphQL schema in Dgraph. Use the /admin API to add a GraphQL schema',
                'Dgraph',
            ]),
    }



# github_directory: dotansimha/graphql-yoga, stars: 7041, last_update: 2022-07-10
from graphdna.detectors.checkers import in_response_text
from graphdna.entities.interfaces.heuristics import IGQLQuery


class GraphQLYoga(IGQLQuery):

    score_factor = 0.62
    genetics = {
        'subscription {__typename }': in_response_text([
            'asyncExecutionResult[Symbol.asyncIterator] is not a function',
            'Unexpected error.',
        ])
    }



# github_directory: neomatrixcode/Diana.jl, stars: 104, last_update: 2022-07-10
from graphdna.detectors.checkers import in_response_text
from graphdna.entities.interfaces.heuristics import IGQLQuery


class DianaJl(IGQLQuery):

    score_factor = 0.5
    genetics = {
        'query { __typename }': in_response_text('Syntax Error GraphQL request (1:1) Unexpected Name \\"queryy\\"'),
    }



# github_directory: graphql-python/graphene, stars: 7291, last_update: 2022-07-10
from graphdna.detectors.checkers import in_response_text
from graphdna.entities.interfaces.heuristics import IGQLQuery


class Graphene(IGQLQuery):

    score_factor = 0.62
    genetics = {
        'aaa': in_response_text('Syntax Error GraphQL (1:1)'),
    }



# github_directory: graphql-go/graphql, stars: 8636, last_update: 2022-07-10
from graphdna.detectors.checkers import in_response_text
from graphdna.entities.interfaces.heuristics import IGQLQuery


class GraphQLGo(IGQLQuery):

    score_factor = 0.64
    genetics = {
        '': in_response_text('Must provide an operation.'),
        'query  { __typename {}': in_response_text('Unexpected empty IN'),
        'query { __typename }': in_response_text('RootQuery'),
    }



from graphdna.detectors.checkers import has_json_key
from graphdna.entities.interfaces.dna import IRequest
from graphdna.entities.interfaces.heuristics import IWebProperty


class Shopify(IWebProperty):

    score_factor = 3.0
    requests = [(IRequest('%%base_url%%/products.json', method='GET'), has_json_key('products'))]



# github_directory: tartiflette/tartiflette, stars: 823, last_update: 2022-07-10
from graphdna.detectors.checkers import in_response_text
from graphdna.entities.interfaces.heuristics import IGQLQuery


class Tartiflette(IGQLQuery):

    score_factor = 0.51
    genetics = {
        'query @a { __typename }': in_response_text('Unknow Directive < @a >.'),
        'query @skip { __typename }': in_response_text('Unknow Directive < @a >.'),
        'query { gqldna }': in_response_text('Field gqldna doesn\'t exist on Query'),
        'query { __typename @deprecated }': in_response_text('Directive < @deprecated > is not used in a valid location.'),
        'queryy { __typename }': in_response_text('syntax error, unexpected IDENTIFIER'),
    }



# github_directory: wp-graphql/wp-graphql, stars: 3261, last_update: 2022-07-10
from graphdna.detectors.checkers import in_response_text
from graphdna.entities.interfaces.heuristics import IGQLQuery


class WPGraphQL(IGQLQuery):

    score_factor = 0.55
    genetics = {
        '': in_response_text('GraphQL Request must include at least one of those two parameters: \\"query\\" or \\"queryId\\"'),
        'query {alias1$1:__typename}': in_response_text('GraphQL Debug logging is not active. To see debug logs, GRAPHQL_DEBUG must be enabled.'),
    }



import asyncio
import logging
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Union

import aiohttp

from graphdna.entities.engines import GraphQLEngine


class IRequest:

    url: str
    method: str
    kwargs: Dict[str, Any]

    def __init__(
        self,
        url: str,
        method: str,
        kwargs: Dict[str, Any] = None,
    ) -> None:
        self.url = url
        self.method = method or 'GET'
        self.kwargs = kwargs or {}

        self.kwargs['allow_redirects'] = False
        self.kwargs['timeout'] = 3


class IHTTPBucket(ABC):

    _headers: Dict[str, str]
    _store: Dict[str, Union[aiohttp.ClientResponse, asyncio.Task, None]]
    _queue: List[asyncio.Task]

    _url: str
    _base_url: str

    _session: Optional[aiohttp.ClientSession]
    _logger: logging.Logger

    @staticmethod
    def hash(request: IRequest) -> str:
        key = hash(hash(request.url) + hash(request.method))

        data = 0
        for k, v in request.kwargs.items():
            data += hash(f'{k},{str(v)}')

        return str((key + data) & 0xffffffff)

    @abstractmethod
    async def put(
        self,
        req: IRequest,
    ) -> str:
        ...

    @abstractmethod
    def get(
        self,
        key: str,
    ) -> Optional[aiohttp.ClientResponse]:
        ...

    @abstractmethod
    async def send_request(
        self,
        request: IRequest,
        key: str,
    ) -> None:
        ...

    @abstractmethod
    async def consume_bucket(self) -> None:
        ...

    @abstractmethod
    async def open_session(self) -> aiohttp.ClientSession:
        ...

    @abstractmethod
    async def close_session(self) -> None:
        ...


class IGraphDNA(ABC):

    _url: str
    _logger: logging.Logger
    _http_bucket: IHTTPBucket

    @property
    def url(self) -> str:
        return self._url

    @abstractmethod
    async def run(self) -> Optional[GraphQLEngine]:
        ...



# github_directory: strawberry-graphql/strawberry, stars: 2400, last_update: 2022-07-10
from graphdna.detectors.checkers import in_response_text
from graphdna.entities.interfaces.heuristics import IGQLQuery


class Strawberry(IGQLQuery):

    score_factor = 0.54
    genetics = {
        'query @deprecated { __typename }': in_response_text('Directive \'@deprecated\' may not be used on query.'),
    }



# github_directory: ohler55/agoo, stars: 790, last_update: 2022-07-10
from graphdna.detectors import in_section
from graphdna.entities.interfaces.heuristics import IGQLQuery


class Agoo(IGQLQuery):

    score_factor = 0.51
    genetics = {
        'query { zzz }': in_section('code', 'eval error'),
    }



# github_directory: sangria-graphql/sangria, stars: 1895, last_update: 2022-07-10
from graphdna.detectors.checkers import in_response_text
from graphdna.entities.interfaces.heuristics import IGQLQuery


class Sangria(IGQLQuery):

    score_factor = 0.53
    genetics = {
        'queryy { __typename }':
            in_response_text('Syntax error while parsing GraphQL query. Invalid input \\"queryy\\", expected ExecutableDefinition or TypeSystemDefinition'),
    }



"""Managing the CLI flow."""

import argparse
import sys
from typing import List, Optional

from graphdna.dna import detect_engine


def parse_args(args: List[str]) -> argparse.Namespace:
    """Parse the arguments."""

    parser = argparse.ArgumentParser()

    parser.add_argument('-u', '--url', dest='url', type=str, help='The URL of the GraphQL endpoint.')

    return parser.parse_args(args)


def validate_args(args: argparse.Namespace) -> None:
    """Validate arguments namespace.

    Throw:
        ValueError: If the arguments are invalid.
    """

    if not args.url:
        raise ValueError('The URL is required.')


def cli(argv: Optional[List[str]] = None) -> None:
    """CLI entry point."""

    if argv is None:
        argv = sys.argv[1:]

    args = parse_args(argv)
    validate_args(args)

    detect_engine(args.url)



# github_directory: directus/directus, stars: 16802, last_update: 2022-07-10
from graphdna.detectors.checkers import in_section
from graphdna.entities.interfaces.heuristics import IGQLQuery


class Directus(IGQLQuery):

    score_factor = 0.78
    genetics = {
        '': in_section('INVALID_PAYLOAD', 'code'),
    }



import asyncio
import logging
from typing import Dict, Optional

import aiohttp

from graphdna.entities.interfaces.dna import IHTTPBucket, IRequest


class HTTPBucket(IHTTPBucket):

    """Defines a HTTP Bucket, which store procedeed requests.

    The goal is to send requests once.
    """

    def __init__(
        self,
        logger: logging.Logger,
        url: str,
        headers: Optional[Dict[str, str]],
    ) -> None:
        self._store = {}
        self._queue = []
        self._session = None

        self._url = url
        self._base_url = '/'.join(url.split('/')[0:3])
        self._headers = headers or {}
        self._logger = logger

    async def put(
        self,
        req: IRequest,
    ) -> str:
        if not self._session:
            await self.open_session()

        key = self.hash(req)
        if key not in self._store:
            task = asyncio.create_task(self.send_request(req, key))
            self._store[key] = task
            self._queue.append(task)

        return key

    def get(
        self,
        key: str,
    ) -> Optional[aiohttp.ClientResponse]:
        value = self._store.get(key)
        assert isinstance(value, aiohttp.ClientResponse) or value is None
        return value

    async def send_request(
        self,
        request: IRequest,
        key: str,
    ) -> None:
        if not self._session:
            self._session = await self.open_session()

        if '%%base_url%%' in request.url:
            request.url = request.url.replace(
                '%%base_url%%',
                self._base_url,
            )

        if '%%url%%' in request.url:
            request.url = request.url.replace(
                '%%url%%',
                self._url,
            )

        if self._logger.level < logging.DEBUG:
            self._logger.debug(f'[{request.method}] {request.url} {request.kwargs}')

        try:
            self._store[key] = await self._session.request(
                request.method,
                request.url,
                **request.kwargs,
            )
        except asyncio.exceptions.TimeoutError:
            self._logger.debug(f'[{request.method}] {request.url} timed out.')
            self._store[key] = None

    async def consume_bucket(self) -> None:
        self._logger.info(f'Consuming bucket of {len(self._queue)} requests.')

        for task in self._queue:
            await task

    async def open_session(self) -> aiohttp.ClientSession:
        self._session = aiohttp.ClientSession(headers=self._headers)
        return self._session

    async def close_session(self) -> None:
        if not self._session:
            return

        await self._session.close()
        self._session = None



# github_directory: 99designs/gqlgen, stars: 7741, last_update: 2022-07-10
from graphdna.detectors.checkers import in_response_text
from graphdna.entities.interfaces.heuristics import IGQLQuery


class GQLGen(IGQLQuery):

    score_factor = 0.63
    score = 200
    genetics = {
        'query  { __typename {}': in_response_text('GRAPHQL_PARSE_FAILED'),
        'query {enumInInput(input: {enum: INVALID})}': in_response_text('GRAPHQL_VALIDATION_FAILED'),
        'query { alias^_:__typename {}': in_response_text('Expected Name, found <Invalid>')
    }



import functools
import json
from typing import List, Union

import aiohttp


def in_response_text(data: Union[str, List[str]]) -> functools.partial:

    if isinstance(data, str):
        data = [data]

    async def __internal(
        data: List[str],
        response: aiohttp.ClientResponse = None,
    ) -> bool:

        assert response, 'Response is required.'

        response_text = await response.text()
        if not response_text:
            return False
        return any(text in response_text for text in data)

    return functools.partial(__internal, data)


def in_section(
    section: str,
    data: Union[str, List[str]],
) -> functools.partial:

    if isinstance(data, str):
        data = [data]

    async def __internal(
        section: str,
        data: List[str],
        response: aiohttp.ClientResponse = None,
    ) -> bool:

        assert response, 'Response is required.'

        response_json = await response.json()
        if response_json or section not in response_json:
            return False

        section_text = json.dumps(response_json[section])
        return any(text in section_text for text in data)

    return functools.partial(__internal, section, data)


def has_json_key(sections: Union[str, List[str]]) -> functools.partial:

    if isinstance(sections, str):
        sections = [sections]

    async def __internal(
        sections: List[str],
        response: aiohttp.ClientResponse = None,
    ) -> bool:

        assert response, 'Response is required.'

        response_json = await response.json()
        if not response_json:
            return False
        return all(section in response_json for section in sections)

    return functools.partial(__internal, sections)



# github_directory: hasura/graphql-engine, stars: 27374, last_update: 2022-07-10
from graphdna.detectors.checkers import in_response_text
from graphdna.entities.interfaces.heuristics import IGQLQuery


class Hasura(IGQLQuery):

    score_factor = 0.95
    genetics = {
        'query @cached {__typename}': in_response_text('query_root'),
        'query @skip {__typename}': in_response_text('directive \\"skip\\" is not allowed on a query'),
        'query { __schema }': in_response_text('missing selection set for \\"__Schema\\"'),
        'query { aa }': in_response_text('field \\"aaa\\" not found in type: \'query_root\''),
    }



# github_directory: graphql-java/graphql-java, stars: 5539, last_update: 2022-07-10
from graphdna.detectors.checkers import in_response_text
from graphdna.entities.interfaces.heuristics import IGQLQuery


class GraphQLJava(IGQLQuery):

    score_factor = 0.59
    genetics = {
        '':
            in_response_text('Invalid Syntax : offending token \'<EOF>\''),
        'query @aaa@aaa { __typename }':
            in_response_text('Validation error of type DuplicateDirectiveName: Directives must be uniquely named within a location.'),
        'query { __typename }':
            in_response_text('Invalid Syntax : offending token \'queryy\''),
    }



"""Manage engines entities."""

from enum import Enum, unique


@unique
class GraphQLEngine(Enum):

    """Represent GraphQL Engines."""

    AGOO = 'Agoo'
    APOLLO = 'Apollo'
    ARIADNE = 'Ariadne'
    AWSAPPSYNC = 'AWSAppSync'
    DGRAPH = 'DGraph'
    DIANAJL = 'DianaJl'
    DIRECTUS = 'Directus'
    FLUTTER = 'Flutter'
    GRAPHENE = 'Graphene'
    GRAPHQLAPIFORWP = 'GraphQLAPIForWP'
    GRAPHQLGO = 'GraphQLGo'
    GRAPHQLJAVA = 'GraphQLJava'
    GRAPHQLPHP = 'GraphQLPHP'
    GRAPHQLYOGA = 'GraphQLYoga'
    HASURA = 'Hasura'
    HYPERGRAPHQL = 'HyperGraphQL'
    JUNIPER = 'Juniper'
    LIGHTHOUSE = 'Lighthouse'
    QGLGEN = 'GQLGen'
    RUBY = 'Ruby'
    SANGRIA = 'Sangria'
    SHOPIFY = 'Shopify'
    STEPZEN = 'Stepzen'
    STRAWBERRY = 'Strawberry'
    TARTIFLETTE = 'Tartiflette'
    WPGRAPHQL = 'WPGraphQL'



from importlib import import_module
from pathlib import Path
from pkgutil import iter_modules
from types import ModuleType
from typing import List


def find_engine(
    engine: str,
    possibilities: List[str],
) -> str:
    for p in possibilities:
        if engine == p.lower():
            return p

    raise ValueError(f'Unknown engine `{engine}`, possibilities: {possibilities}')


def import_heuristics(
    pkg_file: str,
    pkg_name: str,
) -> List[ModuleType]:
    pkg_path = str(Path(pkg_file).parent.absolute())

    filenames = [filename for _, filename, _ in iter_modules([pkg_path])]

    return [import_module(f'{pkg_name}.{filename}') for filename in filenames]



# github_directory: GraphQLAPI/graphql-api-for-wp, stars: 150, last_update: 2022-07-10
from graphdna.detectors.checkers import in_response_text
from graphdna.entities.interfaces.heuristics import IGQLQuery


class GraphQLAPIForWP(IGQLQuery):

    score_factor = 0.5
    genetics = {
        '': in_response_text('The query in the body is empty'),
        'query @doesnotexist { __typename }': in_response_text('No DirectiveResolver resolves directive with name \'doesnotexist\''),
        'query @skip { __typename }': in_response_text('Argument \'if\' cannot be empty, so directive \'skip\' has been ignored'),
        'query aa#aa { __typename }': in_response_text('Unexpected token \\"END\\"'),
        'query {alias1$1:__typename}': in_response_text('QueryRoot'),
    }



# github_directory: hypergraphql/hypergraphql, stars: 146, last_update: 2022-07-10
from graphdna.detectors.checkers import in_response_text
from graphdna.entities.interfaces.heuristics import IGQLQuery


class HyperGraphQL(IGQLQuery):

    score_factor = 0.5
    genetics = {
        'query {alias1:__typename @deprecated}': in_response_text('Validation error of type UnknownDirective: Unknown directive deprecated @ \'__typename\''),
        'zzz { __typename }': in_response_text('Validation error of type InvalidSyntax: Invalid query syntax.'),
    }



from graphdna.detectors import in_response_text
from graphdna.entities.interfaces.heuristics import IGQLQuery


class AWSAppSync(IGQLQuery):

    score_factor = 1
    genetics = {
        'query @skip { __typename }': in_response_text('MisplacedDirective'),
    }



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['graphdna',
 'graphdna.detectors',
 'graphdna.entities',
 'graphdna.entities.interfaces',
 'graphdna.heuristics',
 'graphdna.heuristics.gql_queries',
 'graphdna.heuristics.web_properties']

package_data = \
{'': ['*']}

install_requires = \
['JSON-log-formatter>=0.5.1,<0.6.0', 'aiohttp[speedups]>=3.8.1,<4.0.0']

entry_points = \
{'console_scripts': ['graphdna = graphdna:cli']}

setup_kwargs = {
    'name': 'graphdna',
    'version': '1.0.2',
    'description': 'Fast and powerful GraphQL engine fingerprinting tool',
    'long_description': '# GraphDNA ![PyPI](https://img.shields.io/pypi/v/GraphDNA) [![CI](https://github.com/Escape-Technologies/GraphDNA/actions/workflows/ci.yaml/badge.svg)](https://github.com/Escape-Technologies/GraphDNA/actions/workflows/ci.yaml) [![CD](https://github.com/Escape-Technologies/GraphDNA/actions/workflows/cd.yaml/badge.svg)](https://github.com/Escape-Technologies/GraphDNA/actions/workflows/cd.yaml)\n\nGraphDNA is a tool that uses multiple heuristics to fingerprint GraphQL endpoints.\n\n![Banner](docs/banner.png)\n\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/GraphDNA)\n![PyPI - Downloads](https://img.shields.io/pypi/dm/GraphDNA)\n\n## Getting Started\n\nIt takes only two simple steps to fingerprint an endpoint using GraphDNA.\n\n```bash\npip install graphdna\ngraphdna -u https://example.com/graphql\n```\n\nThe full list of supported engines is accessible [here](https://github.com/Escape-Technologies/GraphDNA/blob/main/graphdna/entities/engines.py).\n\n![Banner](docs/hackerone.png)\n\n## Documentation\n\n```python\nimport logging\nfrom typing import Dict, Optional\n\nfrom graphdna import detect_engine, detect_engine_async\nfrom graphdna.entities import GraphQLEngine\n\ndef detect_engine(\n    url: str,\n    headers: Optional[Dict[str, str]] = None,\n    logger: Optional[logging.Logger] = None,\n) -> Optional[GraphQLEngine]:\n    ...\n\n\nasync def detect_engine_async(\n    url: str,\n    headers: Optional[Dict[str, str]] = None,\n    logger: Optional[logging.Logger] = None,\n) -> Optional[GraphQLEngine]:\n    ...\n```\n\n## Local installation\n\n```bash\ngit clone git@github.com:Escape-Technologies/GraphDNA.git\ncd GraphDNA\nchmod +x ./install-dev.sh\n./install-dev.sh\n```\n\n## Credits\n\n* [Graphw00f](https://github.com/dolevf/graphw00f)\n* [Dolev Farhi](https://github.com/dolevf)\n\n## Contributing\n\nPull requests are welcome. For major changes, please open an issue first to discuss what you would like to change.\n\nPlease make sure to update tests as appropriate.\n\n## License ![PyPI - License](https://img.shields.io/pypi/l/GraphDNA)\n\n[MIT](https://choosealicense.com/licenses/mit/)\n',
    'author': 'Escape Technologies SAS',
    'author_email': 'ping@escape.tech',
    'maintainer': 'Swan',
    'maintainer_email': 'swan@escape.tech',
    'url': 'https://escape.tech/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<=3.11',
}


setup(**setup_kwargs)



# github_directory: nuwave/lighthouse, stars: 2920, last_update: 2022-07-10
from graphdna.detectors.checkers import in_response_text, in_section
from graphdna.entities.interfaces.heuristics import IGQLQuery


class Lighthouse(IGQLQuery):

    score_factor = 0.55
    genetics = {
        'query {__typename @include(if: falsee)}': [
            in_response_text('Internal server error'),
            in_section('category', 'internal'),
        ],
    }



# github_directory: graphql-rust/juniper, stars: 4595, last_update: 2022-07-10
from graphdna.detectors.checkers import in_response_text
from graphdna.entities.interfaces.heuristics import IGQLQuery


class Juniper(IGQLQuery):

    score_factor = 0.58
    genetics = {
        '': in_response_text('Unexpected end of input'),
        'queryy { __typename }': in_response_text('Unexpected \\"queryy\\"'),
    }



# github_directory: mirumee/ariadne, stars: 1789, last_update: 2022-07-10
from graphdna.detectors import in_response_text
from graphdna.entities.interfaces.heuristics import IGQLQuery


class Ariadne(IGQLQuery):

    score_factor = 0.53
    genetics = {
        '': in_response_text('The query must be a string.'),
        'query { __typename @abc }': in_response_text('Unknown directive \'@abc\'.'),
    }



# github_directory: webonyx/graphql-php, stars: 4339, last_update: 2022-07-10
from graphdna.detectors.checkers import in_response_text
from graphdna.entities.interfaces.heuristics import IGQLQuery


class GraphQLPHP(IGQLQuery):

    score_factor = 0.5700000000000001
    genetics = {
        'query ! {__typename}': in_response_text('Syntax Error: Cannot parse the unexpected character \\"?\\".'),
        'query @deprecated {__typename}': in_response_text('Directive \\"deprecated\\" may not be used on \\"QUERY\\".'),
    }



# github_directory: rmosolgo/graphql-ruby, stars: 5088, last_update: 2022-07-10
from graphdna.detectors.checkers import in_response_text
from graphdna.entities.interfaces.heuristics import IGQLQuery


class Ruby(IGQLQuery):

    score_factor = 0.58
    genetics = {
        'query @deprecated { __typename }': in_response_text('\'@deprecated\' can\'t be applied to queries'),
        'query @skip { __typename }': in_response_text('\'@skip\' can\'t be applied to queries (allowed: fields, fragment spreads, inline fragments)'),
        'query { __typename @skip }': in_response_text('Directive \'skip\' is missing required arguments: if'),
        'query { __typename {}': in_response_text('Parse error on \\"}\\" (RCURLY)')
    }



# github_directory: apollographql/apollo-server, stars: 12730, last_update: 2022-07-10
from graphdna.detectors import in_response_text
from graphdna.entities.interfaces.heuristics import IGQLQuery


class Apollo(IGQLQuery):

    score_factor = 0.71
    genetics = {
        'query @deprecated { __typename }':
            in_response_text([
                'Directive \\\"@deprecated\\\" may not be used on QUERY.',
                'Directive \\\"deprecated\\\" may not be used on QUERY.',
            ]),
        'query @skip { __typename }':
            in_response_text([
                'Directive \\\"@skip\\\" argument \\\"if\\\" of type \\\"Boolean!\\\" is required, but it was not provided',
                'Directive \\\"skip\\\" argument \\\"if\\\" of type \\\"Boolean!\\\" is required, but it was not provided'
            ])
    }



"""Manage logging flow."""

import logging
import os
from typing import Optional

import json_log_formatter  # type: ignore[import]


class JSONFormat(json_log_formatter.JSONFormatter):

    """Overload the json formater."""

    def json_record(self, message, extra, record):  # type: ignore[no-untyped-def]
        """Add level and environment to the log record."""

        extra['level'] = (record.levelname or '').lower()
        extra['pathname'] = record.pathname or ''
        extra['lineno'] = record.lineno or ''

        return super().json_record(message, extra, record)


def install_logger(logger: logging.Logger) -> None:
    """Install logger."""

    handlers = {
        'console': logging.StreamHandler(),
        'json': logging.StreamHandler(),
    }
    handlers['json'].setFormatter(JSONFormat())
    handlers['console'].setFormatter(logging.Formatter('%(asctime)s %(levelname)s %(name)s - %(message)s'))

    log_format = os.getenv('LOG_FORMAT', 'console')

    logger.addHandler(handlers[log_format])
    logger.setLevel(logging.DEBUG)

    # Ignore asyncio debug logs
    logging.getLogger('asyncio').setLevel(logging.ERROR)


def setup_logger(name: Optional[str] = None) -> logging.Logger:
    """Setup logger."""

    name = name or 'graphdna'
    logger = logging.getLogger(name)

    if not logger.hasHandlers():
        install_logger(logger)

    return logger



"""Manage heuristics flow."""

import logging
from typing import Optional

from graphdna.entities.engines import GraphQLEngine
from graphdna.entities.interfaces.dna import IHTTPBucket
from graphdna.entities.interfaces.heuristics import IHeuristicsManager
from graphdna.heuristics.gql_queries import GQLQueriesManager
from graphdna.heuristics.web_properties import WebPropertiesManager


class HeuristicsManager(IHeuristicsManager):

    def __init__(
        self,
        logger: logging.Logger,
    ) -> None:
        self._logger = logger
        self._candidates = {}

        self._gql_queries_manager = GQLQueriesManager(self._logger)
        self._web_properties_manager = WebPropertiesManager(self._logger)

    def load(self) -> None:
        self._gql_queries_manager.load()
        self._web_properties_manager.load()

    async def enqueue_requests(
        self,
        bucket: IHTTPBucket,
    ) -> None:
        await self._gql_queries_manager.enqueue_requests(bucket)
        await self._web_properties_manager.enqueue_requests(bucket)

    async def parse_requests(
        self,
        bucket: IHTTPBucket,
    ) -> None:
        async for match, engine in self._gql_queries_manager.parse_requests(bucket):
            self.add_score(match, engine)
        async for match, engine in self._web_properties_manager.parse_requests(bucket):
            self.add_score(match, engine)

    def add_score(
        self,
        cls: object,
        engine: GraphQLEngine,
    ) -> None:
        if engine not in self._candidates:
            self._candidates[engine] = 0

        self._candidates[engine] += cls.score * cls.score_factor

    def display_results(self) -> None:
        self._logger.debug('Pushing heuristics results...')

        for engine, score in self._candidates.items():
            self._logger.debug(f'{engine.name.capitalize()}: {round(score, 2)} pts')

    @property
    def best_candidate(self) -> Optional[GraphQLEngine]:
        """Fetch the best candidate engine.

        If any, the highest confidence will be returned.
        """

        sorted_candidates = sorted(
            self._candidates,
            reverse=True,
            key=lambda x: self._candidates[x],
        )
        candidate = sorted_candidates[0] if self._candidates else None
        candidate_literal = candidate.value if candidate else 'None, are you sure this is a GraphQL endpoint?'
        self._logger.info(f'Best candidate: {candidate_literal}')

        return candidate
